/*
 * Copyright 2006-2007, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "fsstats.h"
#include "log.h"

#include <alloca.h>
#include <assert.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <limits.h>
#include <link.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/statvfs.h>

unsigned long long FsFree = 0, FsSize = 0, Zsize = 0, Zfree = 0, Zpools = 0;

static fsstat_t *Fs = 0;

typedef struct libzfs_handle libzfs_handle_t;
typedef struct zpool_handle zpool_handle_t;
typedef int (*zpool_iter_f)(zpool_handle_t *, void *);

static void *Libzfs = 0;
static libzfs_handle_t *libzfs_handle = 0;
static libzfs_handle_t *(*libzfs_init)(void) = 0;
static int (*zpool_iter)(zpool_iter_f,void *) = 0;
static int (*zpool_iter2)(libzfs_handle_t *, zpool_iter_f, void *) = 0;
static unsigned long long (*zpool_get_space_total)(zpool_handle_t *) = 0;
static unsigned long long (*zpool_get_space_used)(zpool_handle_t *) = 0;
static char *(*zpool_get_name)(zpool_handle_t *) = 0;
static int (*zpool_get_root)(zpool_handle_t *, char *, size_t) = 0;



void init_zfs(void)
{
	Libzfs = dlopen("libzfs.so",RTLD_NOW|RTLD_GLOBAL);
	if (0 == Libzfs)
		return;
	libzfs_init = (libzfs_handle_t *(*)())dlsym(Libzfs,"libzfs_init");
	if (libzfs_init) {
		libzfs_handle = libzfs_init();
		zpool_iter2 = (int (*)())dlsym(Libzfs,"zpool_iter");
		assert(zpool_iter2);
	} else {
		zpool_iter = (int (*)())dlsym(Libzfs,"zpool_iter");
		assert(zpool_iter);
	}
	zpool_get_space_total = (unsigned long long (*)()) dlsym(Libzfs,"zpool_get_space_total");
	assert(zpool_get_space_total);
	zpool_get_space_used = (unsigned long long (*)()) dlsym(Libzfs,"zpool_get_space_used");
	assert(zpool_get_space_used);
	zpool_get_name = (char * (*)()) dlsym(Libzfs,"zpool_get_name");
	assert(zpool_get_name);
	zpool_get_root = (int (*)()) dlsym(Libzfs,"zpool_get_root");
	assert(zpool_get_root);
}


static fsstat_t *
add_fs(const char *dev, const char *mnt, const char *typ,
	unsigned long long blocks, unsigned long long free, unsigned long id)
{
	fsstat_t *f = Fs;

	while (f && strcmp(dev,f->dev))
		f = f->next;
	if (f) {
		strcpy(f->typ,typ);
		f->blocks = blocks;
		f->free = free;
	} else {
		f = malloc(sizeof(fsstat_t));
		bzero(f,sizeof(fsstat_t));
		f->next = Fs;
		f->dev = strdup(dev);
		f->mnt = strdup(mnt);
		f->fsid = id;
		strcpy(f->typ,typ);
		f->blocks = blocks;
		f->free = free;
		f->ops_rate = -1;
		Fs = f;
	}
	return f;
}


static int
analyze_zpool(zpool_handle_t *zh, void *ignored)
{
	unsigned long long s, f;
	char root[PATH_MAX], *name;
	struct statvfs st;
	int err,l;
	fsstat_t *fs, *z;
	
	assert(ignored == 0);
	s = zpool_get_space_total(zh) >> 10;
	f = s - (zpool_get_space_used(zh) >> 10);
	name = zpool_get_name(zh);
	l = strlen(name);
	zpool_get_root(zh,root,PATH_MAX);
	if (root[0] == 0) {
		root[0] = '/';
		strcpy(root+1,name);
	}
	err = statvfs(root,&st);
	assert(err != -1);
	z = add_fs(name,root,"zpool",s,f,st.f_fsid);
	fs = Fs;
	while (fs) {
		if ((strncmp(name,fs->dev,l) == 0) && (fs->dev[l] == '/')) {
			fs->zroot = z;
			dbug("associated filesystem %s with pool %s\n",fs->dev,name);
		}
		fs = fs->next;
	}
	return 0;
}


static void
get_zfs_stats(void)
{
	int err;
	Zsize = Zfree = Zpools = 0;
	if (libzfs_init) {
		err = zpool_iter2(libzfs_handle,analyze_zpool,0);
	} else {
		err = zpool_iter(analyze_zpool,0);
	}
	assert(err == 0);
}


static int
Crit(fsstat_t *a, fsstat_t *b)
{
	if (a->free == b->free)
		return strcmp(a->dev,b->dev) < 0;
	return a->free > b->free;
}


static void
sort_fs(void)
{
	fsstat_t *f = Fs, *h = 0, *t = 0;
	while (f) {
		fsstat_t *n = f->next, *i = h;
		if (f->blocks == 0) {
			free(f->dev);
			free(f->mnt);
			free(f);
			f = n;
			continue;
		}
		while (i && Crit(f,i)
			&& i->next && Crit(f,i->next)) {
			assert(i != i->next);
			i = i->next;
		}
		if (i == 0) {
			if (t) {
				assert(h && !Crit(t, i));
				t->next = f;
				t = f;
			} else {
				assert(h == 0);
				h = t = f;
			}
			f->next = 0;
		} else {
			if (Crit(f,i)) {
				f->next = i->next;
				i->next = f;
			} else {
				f->next = i;
				h = f;
			}
			assert(i != i->next);
		}
		f = n;
		assert(f == 0 || f != f->next);
	}
	Fs = h;
}


int
update_fs_stats(void)
{
	static int numnfs = 0;
	static FILE *mnttab = 0;
	struct statvfs vst;
	fsstat_t *f = Fs;
	char device[PATH_MAX], mnt_pnt[PATH_MAX],
	     tmp_mnt[PATH_MAX] = "", *t = tmp_mnt,
	     fstyp[16];	/* Is sizeof fstyp OK? */
	int nfscount = 0, tmpfs = 1;
	unsigned long tmpfs_id = 0;
	unsigned long long tmpfs_s = 0,tmpfs_u = 0;

	FsFree = 0;
	FsSize = 0;
	while (f) {
		f->blocks = 0;	/* marks fs that have been umounted for
				   removal */
		f = f->next;
	}
	if (mnttab) {
		int ret = fseek(mnttab,0,SEEK_SET);
		assert(ret == 0);
	} else {
		mnttab = fopen("/etc/mnttab","r");
		assert(mnttab != 0);
	}
	while (3 == fscanf(mnttab,"%[^\t]\t%[^\t]\t%[^\t]\t%*[^\n]\n",
				device, mnt_pnt, fstyp)) {
		if (0 == strcmp("nfs",fstyp)) {
			++nfscount;
			continue;
		}
		if ((0 != strcmp("ufs",fstyp)) && (0 != strcmp("pcfs",fstyp))
		   && (0 != strcmp("tmpfs",fstyp)) && (0 != strcmp("vxfs",fstyp)
		   && (0 != strcmp("zfs",fstyp))))
			continue;
		if (-1 == statvfs(mnt_pnt,&vst))
			continue;
		unsigned long long size, avail;
		size = ((unsigned long long) vst.f_blocks * vst.f_frsize) >> 10;
		/*
		   dbug("%s @ %s : %llu/%llu\n",
		   fstyp,
		   mnt_pnt,
		   (unsigned long long)vst.f_bfree,
		   (unsigned long long)vst.f_blocks);
		 */
		avail = ((unsigned long long)vst.f_bavail * vst.f_frsize) >> 10;
		if (0 == strcmp(fstyp,"tmpfs")) {
			tmpfs_s = size;
			tmpfs_u += (size - avail);
			if (t+4 < tmp_mnt+sizeof(tmp_mnt)) {
				if (t != tmp_mnt) {
					*t++ = ',';
					*t++ = ' ';
				}
				t += strlcpy(t,mnt_pnt,tmp_mnt - t + sizeof(tmp_mnt));
			}
			if (tmpfs_id == 0)
				tmpfs_id = vst.f_fsid;
		} else {
			FsSize += size;
			FsFree += avail;
			(void) add_fs(device,mnt_pnt,fstyp,size,avail,vst.f_fsid);
		}
	}
	(void) add_fs("<swap>",tmp_mnt,"tmpfs",tmpfs_s,tmpfs_s - tmpfs_u,tmpfs_id);
	if (Libzfs)
		get_zfs_stats();
	sort_fs();
	if (nfscount != numnfs) {
		numnfs = nfscount;
		return 1;
	}
	return 0;
}


fsstat_t *
get_fs_stats(void)
{
	return Fs;
}

// vim:tw=0
